@@ -343,11 +343,56 @@ gail_widget_get_parent (AtkObject *accessible)
   return parent;
 }
 
-AtkRelationSet*
+static GtkWidget*
+find_label (GtkWidget *widget)
+{
+
+  GList *list, *l;
+  GtkWidget *parent;
+  GtkWidget *label;
+
+  label = NULL;
+  parent = gtk_widget_get_parent (widget);
+  if (GTK_IS_CONTAINER (parent))
+    {
+      list = gtk_container_get_children (GTK_CONTAINER (parent));
+
+      l = list;
+      if (l)
+        {
+          while (l)
+            {
+              if ((GtkWidget *)l->data == widget)
+                break;
+
+              l = l->next;
+            }
+          /*
+           * The widget may claim container as parent but it may not
+           * be listed as a child.
+           */
+          g_return_val_if_fail (l, NULL);
+          if (l->prev && GTK_IS_LABEL (l->prev->data))
+            label = GTK_WIDGET (l->prev->data);
+          else if (l->next && GTK_IS_LABEL (l->next->data))
+            label = GTK_WIDGET (l->next->data);
+          else
+            label = NULL;
+
+          g_list_free (list);
+       }
+    }
+  return label;
+}
+
+static AtkRelationSet*
 gail_widget_ref_relation_set (AtkObject *obj)
 {
   GtkWidget *widget;
   AtkRelationSet *relation_set;
+  GtkWidget *label;
+  AtkObject *array[1];
+  AtkRelation* relation;
 
   g_return_val_if_fail (GAIL_IS_WIDGET (obj), NULL);
 
@@ -368,54 +413,28 @@ gail_widget_ref_relation_set (AtkObject *obj)
        * If the previous or next child of the current widget is a label 
        * and its mnemonic widget is the current widget then set LABELLED_BY 
        * relation on current accessible object.
+       *
+       * If this fails try to find a label at the same level as the widget's
+       * parent. This finds labels for NautilusProgramChooser dialog.
        */
-      GList *list, *l;
-      GtkWidget *parent;
-      GtkWidget *label;
-      gint index;
-      AtkObject *array[1];
-      AtkRelation* relation;
-
-      parent = gtk_widget_get_parent (widget);
-      if (GTK_IS_CONTAINER (parent))
+      label = find_label (widget);
+      if (!label) 
         {
-          list = gtk_container_get_children (GTK_CONTAINER (parent));
+          GtkWidget *parent;
 
-          l = list;
-          if (l)
+          parent = gtk_widget_get_parent (widget);
+          if (parent)
+            label = find_label (parent);
+        }
+      if (label)
+        {
+          if (gtk_label_get_mnemonic_widget (GTK_LABEL (label)) == widget)
             {
-              while (l)
-                {
-                  if ((GtkWidget *)l->data == widget)
-                    break;
+	      array [0] = gtk_widget_get_accessible (label);
 
-                  l = l->next;
-                }
-              /*
-               * The widget may claim container as parent but it may not
-               * be listed as a child.
-               */
-              g_return_val_if_fail (l, relation_set);
-              if (l->prev && GTK_IS_LABEL (l->prev->data))
-                label = GTK_WIDGET (l->prev->data);
-              else if (l->next && GTK_IS_LABEL (l->next->data))
-                label = GTK_WIDGET (l->next->data);
-              else
-                label = NULL;
-
-              g_list_free (list);
-
-              if (label)
-                {
-                  if (gtk_label_get_mnemonic_widget (GTK_LABEL (label)) == widget)
-                    {
-	              array [0] = gtk_widget_get_accessible (label);
-
-	              relation = atk_relation_new (array, 1, ATK_RELATION_LABELLED_BY);
-	              atk_relation_set_add (relation_set, relation);
-	              g_object_unref (relation);
-                    }
-                }
+	      relation = atk_relation_new (array, 1, ATK_RELATION_LABELLED_BY);
+	      atk_relation_set_add (relation_set, relation);
+	      g_object_unref (relation);
             }
         }
     }
@@ -1011,7 +1030,17 @@ static gboolean gail_widget_on_screen (GtkWidget *widget)
           widget->allocation.y + widget->allocation.height <= 0)
         return_value = FALSE;
       else 
-        return_value = TRUE;
+        {
+        /*
+         * Check whether the widget has been placed of the screen. The
+         * widget may be MAPPED as when toolbar items do not fit on the toolbar.
+         */
+          if (widget->allocation.x + widget->allocation.width <= 0 &&
+              widget->allocation.y + widget->allocation.height <= 0)
+            return_value = FALSE;
+          else 
+            return_value = TRUE;
+        }
     }
 
   return return_value;
